import numpy as np
import matplotlib.pyplot as plt


def plot_patterns(X):
    """
    Plot platterns.

    Parameters
    ----------
    X : array
        Data for plotting.
    """
    X = np.atleast_2d(X)
    if X.shape[1] == 1:
        X = X.T
    n_ims, N = X.shape
    sz = int(round(np.sqrt(N)))

    rows = n_ims // 4
    if rows * 4 < n_ims:
        rows += 1
    if n_ims < 4:
        cols = n_ims
    else:
        cols = 4
    fig, axes = plt.subplots(rows, cols, figsize=(cols*2, rows*2))
    if axes.ndim == 1:
        axes = [axes]
    im = 0
    for ax_row in axes:
        for ax in ax_row:
            if im < n_ims:
                ax.imshow(X[im].reshape((sz, sz)), vmin=-1, vmax=1,
                          cmap='gray', interpolation='nearest')
                ax.xaxis.set_visible(False)
                ax.yaxis.set_visible(False)
            else:
                ax.axis('off')
            im += 1
    plt.tight_layout()
    fig.canvas.draw()


class BaseNetwork(object):
    def __init__(self):
        self.T = None

    @classmethod
    def dim_ndim_n_ims_sz_2d(cls, X):
        """
        Get data information.
        """
        ndim = X.ndim
        X = np.atleast_2d(X)
        if X.shape[1] == 1:
            X = X.T
        n_ims, N = X.shape
        sz = int(round(np.sqrt(N)))
        dim = X.shape[-1]
        return dim, ndim, n_ims, sz, X

    @classmethod
    def init_plot_dynamics(cls, X):
        dim, ndim, n_ims, sz, X = cls.dim_ndim_n_ims_sz_2d(X)
        fig, axes = plt.subplots(n_ims, 2, figsize=(2*2, 2*n_ims))
        im_handles = []
        if axes.ndim == 1:
            axes = [axes]
        for im, ax_row in zip(X, axes):
            for ax in ax_row:
                h = ax.imshow(im.reshape((sz, sz)), vmin=-1, vmax=1,
                              cmap='gray', interpolation='nearest')
                ax.xaxis.set_visible(False)
                ax.yaxis.set_visible(False)
            im_handles.append(h)
        axes[0][0].set_title('Orig. pattern')
        axes[0][1].set_title('Updated w/ dynamics')
        plt.tight_layout()
        fig.canvas.draw()
        return fig, im_handles

    @classmethod
    def update_plot_dynamics(cls, fig, im_handles, X):
        dim, ndim, n_ims, sz, X = cls.dim_ndim_n_ims_sz_2d(X)
        for im, h in zip(X, im_handles):
            h.set_data(im.reshape(sz, sz))
        fig.canvas.draw()
        return fig, h

    def run_dynamics(self, X=None, passes=50, display=True):
        """
        Simulates a Hopfield network.

        Parameters
        ----------
        X : array
            Initial input for the network (Nx1), if None, a random start is generated.
        passes : int
            Number of time to iterate over all bits.

        Returns
        -------
        X : array
            the result of running `iters` iterations of the hopfield network
        """
        if self.T is None:
            raise ValueError('net.train(X) needs to be run before net.run_dynamics(X)')

        if X is None:
            X = 2 * np.random.binomial(1, .5, self.T.shape[0]) -1
        dim, ndim, n_ims, sz, X = self.dim_ndim_n_ims_sz_2d(X)
        Y = X.copy()

        # Initialize plot
        if display:
            fig, im_handles = self.init_plot_dynamics(X)

        for ii in range(passes):
            Y_pre = Y.copy()
            for jj in np.random.permutation(dim):

                # Compute net input and flip state accordingly
                Y[:, jj]= 2 * (Y @ self.T[jj, :] > 0) - 1

                if display:
                    # Refresh display
                    self.update_plot_dynamics(fig, im_handles, Y)
            if np.allclose(Y_pre, Y):
                if display:
                    print('Converged at iteraion {} of {}'.format(ii+1, passes))
                break

        if ndim == 1:
            return np.squeeze(X)

        return Y

def corrupt(X, n_bits):
    """
    Corrupts the pattern X by flipping n bits

    Parameters
    ----------
    X : array
        Data to corrupt.
    n : int
        Number of bits to flip per pattern.
    """
    X = X.copy()
    ndim = X.ndim
    X = np.atleast_2d(X)
    if X.shape[1] == 1:
        X = X.T

    for ii in range(X.shape[0]):
        elems = np.random.permutation(X.shape[1])[:n_bits]
        X[ii * np.ones(n_bits, dtype=int), elems] *= -1

    if ndim == 1:
        return np.squeeze(X)
    return X

def generate_patterns(n_patterns, n_bits=100, p=.2):
    """
    Generate random bit vectors.

    Parameters
    ---------
    n_patterns : int
        Number of patterns to generate.
    n_bits : int
        Number of bits per pattern.
    p : float
        Probability of a bit being +1.
    """
    return 2 * np.random.binomial(1, p, size=(n_patterns, n_bits)) - 1

def compare_patterns(original, new):
    """
    Check to see of many of the new patterns are
    equal to any original pattern.

    Parameters
    ---------
    original : array
        Array of original patterns.
    new : array
        Array of converged patterns.
    """
    original = np.atleast_2d(original).astype(float)
    new = np.atleast_2d(new).astype(float)
    original = original/np.linalg.norm(original, axis=-1, keepdims=True)
    new = new/np.linalg.norm(new, axis=-1, keepdims=True)

    aip = abs(new @ original.T).max(axis=1)
    n_good = np.isclose(aip, np.ones_like(aip)).sum()
    print('Of {} converged responses, {} were stored patterns'.format(new.shape[0], n_good))
